import json
import sqlalchemy
import crypto
from db import get_db, User, File
import base64

class RequestHandler(object):
    def __init__(self):
        self.db = get_db()
        self.challenges = {}

    def cleanup(self):
        self.db.close()

    def req_register(self, data):
        try:
            res = self.db.query(User).filter_by(username=data['username'])
            if len(res) > 0:
                raise RuntimeError('Register: username already taken')
        except:
            if not set(['username', 'pubkey']).issubset(set(data.keys())):
                raise RuntimeError('Register: not enough data')
            new_user = User(username=data['username'], pubkey=data['pubkey'], pending_list=json.dumps([]), pending_keys=json.dumps([]))
            self.db.add(new_user)
            try:
                self.db.commit()
                return {'id': new_user.id, 'user_keys': self.get_user_keys()}
            except sqlalchemy.exc.IntegrityError:
                raise RuntimeError('User already exists')

    def req_update(self, data):
            u = self.db.query(User).filter_by(id=data['uid']).first()
            p = u.pending_list
            u.pending_list = json.dumps([])
            k = u.pending_keys
            u.pending_keys = json.dumps([])
            self.db.commit()
            return {'user_keys': self.get_user_keys(), 'pending_list': p, 'pending_keys': k}

    def get_user_info(self, data):
        if 'username' in data.keys():
            user = self.db.query(User).filter_by(username=data['username']).first()
        elif 'id' in data.keys():
            user = self.db.query(User).get(data['id'])
        else:
            raise RuntimeError('User info: not enough data')
        if user:
            return {'username': user.username, 'id': user.id, 'pubkey': user.pubkey, 'pending_list': user.pending_list, 'pending_keys': user.pending_keys}
        else:
            return False

    def get_user_keys(self):
        d = {}
        for user in self.db.query(User).filter():
            d[user.username] = user.pubkey
        return d

    def req_create_file(self, data):
        if not set(['owner_id', 'name', 'name_signature', 'acl',
            'acl_signature', 'data', 'data_signature', 'secret_keys']).issubset(set(data.keys())):
            raise RuntimeError('Create file: not enough data')

        new_file = File(owner_id=data['owner_id'], name=data['name'],
                        name_signature=data['name_signature'],
                        acl=data['acl'],
                        acl_signature=data['acl_signature'],
                        data=data['data'],
                        data_signature=data['data_signature'])

        owner_info = self.get_user_info({'id': data['owner_id']})
        # verify signatures
        if not (crypto.rsa_verify(owner_info['pubkey'], base64.b64decode(new_file.acl_signature), new_file.acl) and \
                crypto.rsa_verify(owner_info['pubkey'], base64.b64decode(new_file.name_signature), base64.b64decode(new_file.name)) and \
                crypto.rsa_verify(owner_info['pubkey'], base64.b64decode(new_file.data_signature), base64.b64decode(new_file.data))):
            raise RuntimeError('Create file: signatures not valid')

        self.db.add(new_file)
        self.db.commit()

        sks = json.loads(data['secret_keys'])

        # add new file to all other users' pending lists
        for u in json.loads(data['acl']).keys():
            if not u == owner_info['username']:
                user = self.db.query(User).filter_by(username=u).first()
                p = json.loads(user.pending_list)
                p.append(new_file.id)
                user.pending_list = json.dumps(p)
                self.db.commit()

                s = json.loads(user.pending_keys)
                s.append(sks[user.username])
                user.pending_keys = json.dumps(s)
                self.db.commit()

        return {'id': new_file.id}

    def req_update_file(self, data):
        if not set(['id', 'data', 'data_signature']).issubset(set(data.keys())):
            raise RuntimeError('Update file: not enough data')

        f = self.db.query(File).get(data['id'])
        user = self.get_user_info({'id': data['uid']})
        if f:
            acl = json.loads(f.acl)
            if not ((user['username'] in acl.keys()) and (acl[user['username']] in ['w', 'rw'])):
                raise RuntimeError("Update file: invalid permissions")
            if not (crypto.rsa_verify(user['pubkey'], base64.b64decode(data['data_signature']), base64.b64decode(data['data']))):
                raise RuntimeError('Update file: signatures not valid')
            f.data = data['data']
            f.data_signature = data['data_signature']
            self.db.commit()
            return {'id': f.id}
        else:
            return {}

    def req_rename_file(self, data):
        if not set(['id', 'name', 'name_signature']).issubset(set(data.keys())):
            raise RuntimeError('Rename file: not enough data')

        f = self.db.query(File).get(data['id'])
        user = self.get_user_info({'id': data['uid']})
        if f:
            acl = json.loads(f.acl)
            if not ((user['username'] in acl.keys()) and (acl[user['username']] in ['w', 'rw'])):
                raise RuntimeError("Rename file: invalid permissions")
            if not (crypto.rsa_verify(user['pubkey'], base64.b64decode(data['name_signature']), base64.b64decode(data['name']))):
                raise RuntimeError('Rename file: signatures not valid')
            f.name = data['name']
            f.name_signature = data['name_signature']
            self.db.commit()
            return {'id': f.id}
        else:
            return {}

    def req_read_file(self, data):
        if 'id' not in data.keys():
            raise RuntimeError('Get file: not enough data')

        f = self.db.query(File).get(data['id'])
        user = self.get_user_info({'id': data['uid']})
        if f:
            acl = json.loads(f.acl)
            if not ((user['username'] in acl.keys()) and (acl[user['username']] in ['r', 'w', 'rw'])):
                raise RuntimeError("Read file: invalid permissions")
            return {'name': f.name, 'name_signature': f.name_signature,
                    'owner': self.get_user_info({'id': f.owner_id})['username'], 'acl': f.acl, 'acl_signature': f.acl_signature,
                    'data': f.data, 'data_signature': f.data_signature}
        else:
            return {'id': -1}

    def req_delete_file(self, data):
        if 'id' not in data.keys():
            raise RuntimeError('Get file: not enough data')

        f = self.db.query(File).get(data['id'])
        user = self.get_user_info({'id': data['uid']})
        if f:
            if user['id'] != f.owner_id:
                # Non owner is deleting
                return {'id': -2}
            self.db.delete(f)
            self.db.commit()
            # Owner is deleting
            return {'id': -1}
        else:
            # File does not exist
            return {'id': -3}
